
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <math.h>
#include "minako.h"

/* *************************************************************** Strukturen */

typedef struct
{
	char* def;
	symtabSymbol* param[20];
	int argCount;
} preprocessorFunction;

/* ********************************************************** Hilfsfunktionen */

static symtabSymbol* newPreprocessorMacro(const char* id, const char* def);
static void preprocessorMacroDeclare(symtab* table, const symtabSymbol* macro);

static symtabSymbol* newPreprocessorFunction(const char* id, const char* def);
static void deletePreprocessorFunction(preprocessorFunction* func);
static void preprocessorFunctionDeclare(symtab* table, const symtabSymbol* symb,
                                        stack* parameter);

static void print(const char* id, const symtabSymbol* sym);

/* ************************************************************ Hauptschleife */

#define STR(MACRO) _STR(MACRO)
#define _STR(MACRO) #MACRO

#define MAX(a,b) ((a)-((((a)-(b))>>((sizeof(a)<<3)-1))&((a)-(b))))

int main()
{
	symtab* table = newSymtab(4);
	stack* param = newStack();
	
	preprocessorMacroDeclare(table, newPreprocessorMacro("INT_MAX", STR(INT_MAX)));
	preprocessorMacroDeclare(table, newPreprocessorMacro("ULONG_MAX", STR(ULONG_MAX)));
	
	printf("Aktuell in der Symboltabelle:\n");
	symtabPrint(table, print);
	printf("\n");
	
	symtabPush(table);
	
	preprocessorMacroDeclare(table, newPreprocessorMacro("PI", STR(M_PI)));
	preprocessorMacroDeclare(table, newPreprocessorMacro("SQRT2", STR(M_SQRT2)));
	preprocessorMacroDeclare(table, newPreprocessorMacro("INT_MAX", "17"));
	
	stackPush(param, newPreprocessorMacro("a", ""));
	stackPush(param, newPreprocessorMacro("b", ""));
	
	preprocessorFunctionDeclare(table, newPreprocessorFunction("MAX", STR(MAX(a,b))), param);
	
	printf("Aktuell in der Symboltabelle:\n");
	symtabPrint(table, print);
	printf("\n");
	
	symtabPop(table);
	
	printf("Aktuell in der Symboltabelle:\n");
	symtabPrint(table, print);
	printf("\n");
	
	deleteSymtab(table);
	
	return 0;
}

/* *********************************************************** Implementation */

#define NEW(_struct) (_struct*)malloc(sizeof(_struct))

/* ****************************************************** Präprozessor-Makros */

static symtabSymbol* newPreprocessorMacro(const char* id, const char* def)
{
	/* erstelle ein neues Symbol der Klasse 'vmcr' (variablenähnliches Makro) */
	symtabSymbol* sym = newSymtabSymbol('vmcr', id);
	
	/* kopiere den Definitionsstring als assoziiertes Datum */
	sym->data = (void*)strdup(def);
	
	/* der Standarddestruktor ist die Funktion free - sie ist für die Freigabe
	 * des Strings absolut ausreichend */
	
	return sym;
}

static void preprocessorMacroDeclare(symtab* table, const symtabSymbol* macro)
{
	assert(table && macro && macro->class == 'vmcr');
	
	if (!symtabDeclare(table, macro))
	{
		fprintf(stderr, "Das Makro %s ist schon deklariert worden.", macro->id);
		exit(-1);
	}
}

/* ******************************************** Präprozessor Makro-Funktionen */

static symtabSymbol* newPreprocessorFunction(const char* id, const char* def)
{
	/* erstelle ein neues Symbol der Klasse 'fmcr' (funktionsähnliches Makro) */
	symtabSymbol* sym = newSymtabSymbol('fmcr', id);
	preprocessorFunction* func = NEW(preprocessorFunction);
	
	/* kopiere den Definitionsstring als assoziiertes Datum */
	sym->data = func;
	sym->destructor = (fDestructor) deletePreprocessorFunction;
	
	func->def = strdup(def);
	func->argCount = 0;
	
	return sym;
}

static void deletePreprocessorFunction(preprocessorFunction* func)
{
	int i;
	
	for (i = 0; i < func->argCount; ++i)
		deleteSymtabSymbol(func->param[i]);
	
	free(func->def);
	free(func);
}

static void preprocessorFunctionDeclare(symtab* table, const symtabSymbol* symb,
                                        stack* parameter)
{
	preprocessorFunction* func;
	
	assert(table && symb->class == 'fmcr');
	
	func = (preprocessorFunction*) symb->data;
	
	if (!symtabDeclare(table, symb))
	{
		fprintf(stderr, "Das Makro %s ist schon deklariert worden.", symb->id);
		exit(-1);
	}
	
	while (!stackIsEmpty(parameter))
	{
		func->param[func->argCount] = (symtabSymbol*) stackPop(parameter);
		
		++func->argCount;
		
		if (func->argCount >= 20)
		{
			fprintf(stderr, "Sorry, nur 20 Parameter sind zugelassen.");
			exit(-1);
		}
	}
	
	deleteStack(parameter);
}

#undef NEW

/* ****************************************************************** Ausgabe */

static void preprocessorMacroPrint(const char* def, const char* id)
{
	if (strcmp(def, ""))
		printf("%s -> %s", id, def);
	else
		printf("%s", id);
}

static void preprocessorFunctionPrint(preprocessorFunction* func, const char* id)
{
	printf("%s(", id);
	
	if (func->argCount)
	{
		int i = func->argCount - 1;
		preprocessorMacroPrint(func->param[i]->data, func->param[i]->id);
		
		while (i--)
		{
			printf(", ");
			preprocessorMacroPrint(func->param[i]->data, func->param[i]->id);
		}
	}
	
	printf(") -> %s", func->def);
}

static void print(const char* id, const symtabSymbol* sym)
{
	switch (sym->class)
	{
	case 'vmcr':
		preprocessorMacroPrint(sym->data, id);
		printf("\n");
		break;
		
	case 'fmcr':
		preprocessorFunctionPrint(sym->data, id);
		printf("\n");
		break;
	}
}
